/* Implementation of common functions
 *
 * File: parapov.cpp		Author: Matthias Fueller
 * Date: 13.11.2002 		Version: 0.1
 *
 * Change log:
 */
#include <stdio.h>
#include <stdlib.h>
#include "frame.h"
#include "povray.h"
#include "userio.h"
#include "optout.h"
#include "parapov.h"


/* Global variables */
int mytid;
int ntasks;

/* MPI datatypes */
MPI_Datatype MPI_work_t;
MPI_Datatype MPI_error_t;
MPI_Datatype MPI_info_t;
MPI_Datatype MPI_stat_t;


/* static functions */
static void create_MPI_types ();
static void para_init_options ();


/* para_init
 *
 * Initializes MPI
 */
int para_init (int* argc, char** argv[])
{
  int err;

  err = MPI_Init (argc, argv);
  if (err != MPI_SUCCESS)
  {
    para_mpi_error ("failed to init MPI", err);
    return 0;
  }

  create_MPI_types ();

  MPI_Comm_rank (MPI_COMM_WORLD, &mytid);
  MPI_Comm_size (MPI_COMM_WORLD, &ntasks);
#ifdef PARA_DEBUG
  stream_printf(DEBUG_STREAM, "mytid %d ntasks %d\n", mytid, ntasks);
#endif

  if ((mytid < 0) || (ntasks < 1))	/* valid values? 		*/
  {
    stream_printf (FATAL_STREAM, "Failed to init PARAPOV\n");
    return 0;
  }

  err = MPI_Barrier (MPI_COMM_WORLD);	/* synchronize all tasks 	*/
  if (err != MPI_SUCCESS)
    return 0;
    
  para_init_options ();

  return 1;
}


/* para_close
 *
 * Cleans up.
 */
void para_close ()
{
  MPI_Finalize ();
}


/* para_mpi_error
 *
 * Writes an MPI error message to the FATAL_STREAM of povray
 */
void para_mpi_error (const char* msg, int err)
{
  char buf[ERR_BUF_SIZE];
  int len;

  if (err == MPI_SUCCESS)
    return;

  len = ERR_BUF_SIZE;
  MPI_Error_string (err, buf, &len);

  stream_printf (FATAL_STREAM, "MPI Error: %s, %s\n", msg, buf);
}


/* create_MPI_types
 *
 * Builds the MPI structure types needed for communication
 */
static void create_MPI_types ()
{
  int 		work_block_counts [1] = { 6 };
  MPI_Datatype  work_types [1] = { MPI_INT };
  MPI_Aint	work_offsets [1] = { 0 };
  int 		err_block_counts [2] = { 1, ERR_BUF_SIZE };
  MPI_Datatype  err_types [2] = { MPI_INT, MPI_CHAR };
  MPI_Aint	err_offsets [2];
  int		info_block_counts [1] = { INFO_NAME_LEN * 4 };
  MPI_Datatype	info_types [1] = { MPI_CHAR };
  MPI_Aint	info_offsets [1] = { 0 };
  int		stat_block_counts [3] =
  		{ 1, 1, MaxStat * 2 };
  MPI_Datatype 	stat_types [3] = { MPI_INT, MPI_DOUBLE, MPI_LONG };
  MPI_Aint	stat_offsets [3];

  struct 	error_s err;
  struct 	stat_s	 stat;
  int		a;

  MPI_Get_address (&err.mpi_error, &err_offsets [0]);
  MPI_Get_address (&err.msg, &err_offsets [1]);
  for (a = 1; a >= 0; a--)
  {
    err_offsets [a] -= err_offsets [0];
  };

  MPI_Get_address (&stat.frame, &stat_offsets [0]);  
  MPI_Get_address (&stat.trender, &stat_offsets [1]);
  MPI_Get_address (&stat.stats, &stat_offsets [2]);
  for (a = 2; a >= 0; a--)
  {
    stat_offsets [a] -= stat_offsets [0];
  };

  MPI_Type_create_struct (1, work_block_counts, work_offsets,
			  work_types, &MPI_work_t);
  MPI_Type_commit (&MPI_work_t);

  MPI_Type_create_struct (2, err_block_counts, err_offsets,
  			  err_types, &MPI_error_t);
  MPI_Type_commit (&MPI_error_t);

  MPI_Type_create_struct (1, info_block_counts, info_offsets,
  			  info_types, &MPI_info_t);
  MPI_Type_commit (&MPI_info_t);

  MPI_Type_create_struct (3, stat_block_counts, stat_offsets,
  			  stat_types, &MPI_stat_t);
  MPI_Type_commit (&MPI_stat_t);
}


/* para_init_options
 *
 * sets the default values for the parapov options
 */
void para_init_options ()
{
  opts.Para_Chunk_Width = DEFAULT_CHUNK_WIDTH;
  opts.Para_Chunk_Height = DEFAULT_CHUNK_HEIGHT;
  opts.Para_Slave_Niceness = DEFAULT_NICENESS;
  opts.Para_Slave_Logging = DEFAULT_LOGGING;
}


/* print_parapov_credits
 *
 * prints information about the parapov authors
 */
void print_parapov_credits ()
{
  Banner ("\nThis is Parapov V1.0, a parallel extension to povray");
  Banner ("\n");
}
