/* Everything that is used by the client and by the server
 *
 * File: parapov.h		Author: Matthias Fueller
 * Date: 13.11.2002 		Version: 0.1
 *
 * Change log:
 *	12.12.2002 - added default values for options
 */
#ifndef __PARAPOV_H__
#define __PARAPOV_H__


#include <mpi.h>			/* for MPI_Datatype 		*/
#include "povray.h"			/* for MaxStat 			*/


/* Global constants */
#define ERR_BUF_SIZE	512   		/* buf size in struct err_s	*/
#define INFO_NAME_LEN	65		/* name len in info_s		*/

/* Default values for options */
#define DEFAULT_CHUNK_WIDTH	32
#define DEFAULT_CHUNK_HEIGHT	32
#define DEFAULT_NICENESS	5
#define DEFAULT_LOGGING		1


/* Useful macros */
#define IS_PARA_SLAVE		(mytid > 0)
#define IS_PARA_MASTER		(mytid == 0)
#define IS_PARA_REAL_MASTER	((mytid == 0) && (ntasks > 1))
#define IS_PARA_WORKER		((mytid > 0) || (ntasks == 1))


/* Communication codes */
enum
{
  PARA_SLAVE_CONNECTED = 1,		/* new slave connected 		*/
  PARA_SLAVE_EXITED,			/* slave has terminated 	*/
  PARA_KILL_SLAVE,			/* tells slave to exit 		*/
  PARA_REQUEST_WORK,			/* slave requests work 		*/
  PARA_SEND_WORK,			/* send work to the slave 	*/
  PARA_SEND_STATS,			/* send stats to the master 	*/
  PARA_SEND_RESULTS,			/* send results to the master 	*/
  PARA_ERROR				/* error in slave has occured 	*/
};


/*
 * MPI Types and their C structures
 */

/*
 * Work structure, used for assigning work and sending the results
 */
struct work_s
{
  int frame;				/* frame number			*/
  int block_num;			/* block number			*/
  int left,                        	/* first column of block	*/
      up,				/* first row of block		*/
      right,				/* last column of block		*/
      down;				/* last row of block		*/
};
extern MPI_Datatype MPI_work_t;


/*
 * Error structure, used for sending an error msg to the master
 */
struct error_s
{
  int  mpi_error;			/* the mpi error - optional	*/
  char msg [ERR_BUF_SIZE];		/* error msg			*/
};
extern MPI_Datatype MPI_error_t;


/*
 * Info structure - some Info about the slave (system etc)
 */
struct info_s
{
  char sysname [INFO_NAME_LEN];		/* os name			*/
  char nodename [INFO_NAME_LEN];	/* name of the slave		*/
  char machine [INFO_NAME_LEN];		/* machine type			*/
  char logname [INFO_NAME_LEN];		/* name of the log file		*/
};
extern MPI_Datatype MPI_info_t;


/*
 * Statistics structure - contains statistice about the rendering
 * and the time consumed for one frame. The slave sends this to
 * the master at the end of each frame
 */
struct stat_s
{
  int 	  frame;			/* frame number			*/
  double  trender;			/* render time for this frame	*/
  COUNTER stats [MaxStat];		/* povray statistics		*/
};
extern MPI_Datatype MPI_stat_t;

/* Global variables */
extern int mytid;			/* my MPI task id 		*/
extern int ntasks;			/* number of MPI tasks 		*/


/* Functions */
int  para_init (int* argc, char** argv[]);
void para_close ();
void para_mpi_error (const char* msg, int err);
void print_parapov_credits ();


#endif /* __PARAPOV_H__  */
