/* Server interface implementation of parapov.
 * 
 * This modul contains all the things done by the parapov server.
 * The server holds status information about each client, frame,
 * and block. He tiles each frame in blocks. The block size may be
 * spezified by the user. Each block is then send to a client
 * as a work_s struct. After a client finishes rendering a block
 * the server collects the work results und puts them together.
 * He displays rendered lines in the status window and
 * writes them out into the output file. The master also collects
 * status information from each client about the rendering.
 * After all the work is done, he kills the client tasks and
 * prints out some status information about the rendering, 
 * info about the parallel parts and the performance of each client. 
 * At least the server exits ... ;D
 *
 * Datei: parapov_mas.cpp	Autor: Benjamin Goldbach
 * Datum: 14.11.2002 		Version: 0.1
 *
 * Historie:
 *
 */


#include <assert.h>
#include "frame.h"
#include "pov_util.h"
#include "optout.h"
#include "colutils.h"
#include "renderio.h"
#include "povproto.h"
#include "statspov.h"
#include "parapov.h"
#include "parapov_mas.h"


/* Global constatns */
#define RENDER_STATS_FORMAT	0		/* Flag for para_print_stats ()	*/
#define TOTAL_STATS_FORMAT	1		/* Flag for para_print_stats ()	*/


/* static variables */
static MPI_Status status;		/* needet for MPI operations	*/
static int task;			/* the current slave task	*/
static int para_chunk_height;		/* height of a block		*/
static int para_chunk_width;		/* width of a block		*/
static int grid_points;			/* number of blocks		*/
static int grid_ppline;			/* number of blocks in a row	*/
static int npixels_done = 0;		/* pixels already rendered	*/
static int frames;			/* number of frames to do	*/
static int frame = -1;			/* the current frame		*/
static int next_frame_finish;		/* next frame to render		*/
static int frame_display;		/* the frame that is displayed  */
static para_frame_stat* frame_stat;	/* Info on the frames		*/
static para_slave_stat* slave_stat;	/* Info on the slaves		*/
static int grid_lines_done = 0;		/* number of block lines done	*/
static COUNTER* statistics;		/* to store client render stats	*/

/* static functions */
static void para_master_exit ();
static int  para_start_frame ();
static void para_finish_frame ();
static void para_send_work (int block);
static void para_assign_work();
static void para_display_plot (COLOUR* row_ptr, int row, int col,
                               int ncols);
static void para_receive_data ();
static void para_write ();
static void para_dump_time (int flag);
static void para_receive_stats ();
static void para_print_stats ();


/* para_master_exit
 *
 * Does some cleanup, prints out parapovs status 
 * information and exits.
 */
static void para_master_exit ()
{
  int flag,				/* flag for MPI_Iprobe		*/
      err;				/* slave exit code		*/

  POV_DISPLAY_CLOSE (0);

  /* Get all pending PARA_REQUEST_WORK messages
   */
  do
  {
    MPI_Iprobe (MPI_ANY_SOURCE, PARA_REQUEST_WORK, MPI_COMM_WORLD,
                &flag, &status);
    task = status.MPI_SOURCE;
    if (flag)
    {
      MPI_Recv ((char*) NULL, 0, MPI_CHAR, task, PARA_REQUEST_WORK,
                MPI_COMM_WORLD, &status);
    };
  } while (flag);

  /* kill all the slave tasks, to be sure */
  para_master_kill_slaves ();

  /* Synchronize with the slave. We must wait until the slaves have
   * send their statistics. The corresponding MPI_Barrier in the
   * slave code is in para_slave_exit.
   * This may be bad because one blocked slave could block the
   * whole thing, another way would be to use a simple timeout,
   * like sleep (1)
   */
  MPI_Barrier (MPI_COMM_WORLD);

  /* collect stats from slaves and process pending messages */
  do
  {
    MPI_Iprobe (MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD,
    		&flag, &status);
    task = status.MPI_SOURCE;

    if(flag)
    {
      switch (status.MPI_TAG)
      {
        case PARA_SEND_STATS:
#ifdef PARA_DEBUG
          Debug_Info ("Receiving stats from task %d...\n", task);
#endif
          para_receive_stats ();
	break;
        case PARA_SLAVE_EXITED:
          MPI_Recv (&err, 1, MPI_INT, task,
                    PARA_SLAVE_EXITED, MPI_COMM_WORLD, &status);
#ifdef PARA_DEBUG
          Debug_Info ("slave %d exited with\n");
#endif
        break;
        default:
#ifdef PARA_DEBUG
          Debug_Info ("unknown message: %d from task %d\n",
          	      status.MPI_TAG, task);
#endif
          flag = 0;			/* leave while loop		*/
      };
    };
  } while (flag);

  /* Synchronize with slaves. This prevents the slave from exiting
   * before we have his statistics. If the slave exits too early,
   * there can be problems receiving his messages. The corresponding
   * MPI_Barrier in the slave code is in para_slave_exit
   */
  MPI_Barrier (MPI_COMM_WORLD);

  para_close ();			/* finalize MPI 		*/

  para_print_stats ();			/* print parapovs status info	*/
  exit (0);
}


/* para_start_frame
 *
 * Do all the stuff, that is necessary for a new frame
 * (allocate memory, do a shellout, open the output file)
 * This is called before the work is sent.
 *
 * A lot of this is taken from 'FrameRender'
 *
 * returns:	0	.. OK send the work
 *		1	.. Shellout gave SKIP_ONCE_RET
 *			-> this frame is done
 */
static int para_start_frame ()
{
  int i;
  int Frame_Result;

  if (frame_stat[frame].status != PARA_FRAME_FREE)
  {
    Warning (0, "Frame %d already started.\n", frame);
    return 1;
  };

  if (opts.Options & VERBOSE)
  {
    Status_Info ("\nStarting frame %d...", opts.FrameSeq.InitialFrame
                 + frame);
  };

  frame_stat[frame].status = PARA_FRAME_WORKING;

  /* allocate row pointers */
  frame_stat[frame].row_ptr = (COLOUR **) POV_CALLOC (opts.Last_Line,
      sizeof (COLOUR *), "frame_stat.row_ptr");

  opts.FrameSeq.FrameNumber = opts.FrameSeq.InitialFrame + frame;
  setup_output_file_name ();

  /* Execute a shell-out command before tracing */
  Frame_Result = POV_SHELLOUT (PRE_FRAME_SHL);

  if (Frame_Result == ALL_SKIP_RET)
  {
    /* stop completely */
    for (i = 0; i < frames; i++)
    {
      frame_stat[i].block_to_assign = grid_points;
      frame_stat[i].status = PARA_FRAME_DONE_SHELLOUT;
    };
    /* continue searching, frames before this one, should be finished */
    return 1;
  };

  /* open output file */
  init_output_file_handle ();
  open_output_file ();
  Initialize_Renderer ();

  /* Start displaying */
  if ((opts.Options & DISPLAY) && !Display_Started)
  {
    Status_Info ("\nDisplaying frame %d...", opts.FrameSeq.InitialFrame
                 + frame);
    POV_DISPLAY_INIT (0, Frame.Screen_Width, Frame.Screen_Height);
    Display_Started = 1;
    frame_display = frame;
    Status_Info ("\n");
  };

  /* read already rendered parts */
  if ((opts.Options & DISKWRITE) && (opts.Options & CONTINUE_TRACE))
  {
    int fl = opts.First_Line;		/* remember start of image 	*/

    /* read -> changes opts.First_Line */
    Read_Rendered_Part (Actual_Output_Name);

    frame_stat[frame].row_to_write =
        frame_stat[frame].row_to_malloc = opts.First_Line;

    frame_stat[frame].block_to_write = frame_stat[frame].block_to_assign =
        (opts.First_Line / para_chunk_height) * grid_ppline;
    /* blocks above the new start are already finished */
    for (i = 0; i < frame_stat[frame].block_to_assign; i++)
    {
      frame_stat[frame].block_stat[i].status = PARA_GRID_POINT_WRITTEN;
    };
    if (opts.Options & VERBOSE)
    {
      Status_Info ("Continuing with block %d in frame %d\n",
                   frame_stat[frame].block_to_assign, frame);
    };

    opts.First_Line = fl;		/* set back the old value 	*/

    /* this should not be necessary */
    if (opts.Last_Line > Frame.Screen_Height)
    {
      opts.Last_Line = Frame.Screen_Height;
    };
    if (opts.Last_Column > Frame.Screen_Width)
    {
      opts.Last_Column = Frame.Screen_Width;
    };
  };

  /* make a copy of the Output_File_Handle */
  frame_stat[frame].output_file = Output_File;

  if (opts.Options & VERBOSE)
  {
    Status_Info ("\n");
  };

  return 0;
}


/* para_finish_frame
 *
 * checks, if the next frame in the sequence finished,
 * closes the frame and display if true und does a shellout
 */
static void para_finish_frame ()
{
  /* check, if frame has really finished */
  if (frame_stat[frame].row_to_write < opts.Last_Line)
  {
    return;
  };

  /* if not already closed */
  if (frame_stat[frame].status < PARA_FRAME_DONE)
  {
    if (opts.Options & VERBOSE)
    {
      Status_Info ("\nFinishing frame %d...",
                   /*opts.FrameSeq.InitialFrame*/ + frame);
    };
    Status_Info ("row_to_write: %d\n", frame_stat[frame].row_to_write);

    frame_stat[frame].status = PARA_FRAME_DONE;

    POV_FREE (frame_stat[frame].row_ptr);

    /* Close output file */
    if (frame_stat[frame].output_file)
    {
      POV_DELETE (frame_stat[frame].output_file, Image_File_Class);
    };

    if (frame == frame_display)
    {
      POV_DISPLAY_FINISHED (0);

      if ((opts.Options & DISPLAY) && Display_Started)
      {
        POV_DISPLAY_CLOSE (0);

        Display_Started = 0;
      };

      frame_display = -1;
    };

    if (opts.histogram_on)
    {
      write_histogram (opts.Histogram_File_Name);
    };
  };

  if (frame_stat[frame].status < PARA_FRAME_DONE_SHELLOUT)
  {
    if (frame == next_frame_finish)
    {
      int Frame_Result;

      frame_stat[frame].status = PARA_FRAME_DONE_SHELLOUT;

      opts.FrameSeq.FrameNumber = opts.FrameSeq.InitialFrame + frame;
      setup_output_file_name();

      Frame_Result = POV_SHELLOUT (POST_FRAME_SHL);

      if ((Frame_Result == SKIP_ONCE_RET) ||
          (Frame_Result == ALL_SKIP_RET))
      {
        next_frame_finish = frames;
        return;
      };

      /* one more frame finished */
      next_frame_finish++;

      /* check next frame (might have finished earlier) */
      frame++;
      if (frame < frames)
        para_finish_frame ();
    };
  };

  if (opts.Options & VERBOSE)
  {
    Status_Info ("\n");
  };
}


/* para_send_work
 *
 * creates a work_s struct with the next block 
 * block to render and sends it to the client
 *
 * parameters:	block	number of the block to be send
 */
static void para_send_work (int block)
{
  struct work_s work;

  frame_stat[frame].block_stat[block].status = PARA_GRID_POINT_ASSIGNED;
  frame_stat[frame].block_stat[block].assigned_tid = task;

  work.frame = frame;
  work.block_num = block;

  work.up = (block / grid_ppline) * para_chunk_height + opts.First_Line;
  work.down = (work.up + para_chunk_height);

  if (work.down >= opts.Last_Line)
  {
    work.down = opts.Last_Line;
  };

  work.left = (block % grid_ppline) * para_chunk_width +
              opts.First_Column;
  work.right = (work.left + para_chunk_width);

  if (work.right >= opts.Last_Column)
  {
    work.right = opts.Last_Column;
  };

#ifdef PARA_DEBUG
  Status_Info ("Assigning block: %d, frame %d of %d to task %d = %s.\n",
               block, frame, frames, task,
               slave_stat[task].info.nodename);
  Status_Info ("  block %d: up=%d left=%d down=%d right=%d\n",
          work.block_num, work.up, work.left, work.down, work.right);
#endif

  MPI_Send (&work, 1, MPI_work_t, task,
            PARA_SEND_WORK, MPI_COMM_WORLD);

  frame_stat[frame].block_to_assign++;
}


/* para_assign_work
 *
 * A slave requested work, now find free block
 */
static void para_assign_work()
{
  int min_assigned;
  int found;
  int f;

  /* search for a free block, take a frame with minimal number of
   * assigned blocks
   */
  found = 0;
  min_assigned = grid_points;

  frame = slave_stat[task].frame_assigned;

  /* if slave was already assigned and that frame has still free blocks */
  if ((frame >= 0) &&
      (frame_stat[frame].block_to_assign < grid_points))
  {
    min_assigned = 0;
    found = 1;
  };

  /* find free frame with minimal number of assigned frames */
  for (f = 0; (f < frames) && (min_assigned > 0); f++)
  {
    if ((frame_stat[f].block_to_assign < min_assigned) &&
        (frame_stat[f].block_to_assign < grid_points))
    {
#ifdef PARA_DEBUG
      Status_Info ("\nfound: frame %d, ass: %d, min: %d\n",
                   f, frame_stat[f].block_to_assign, min_assigned);
#endif
      min_assigned = frame_stat[f].block_to_assign;
      frame = f;
      found = 1;
    };
  };

  if (found)
  {
    /* first block for that frame */
    if (frame_stat[frame].block_to_assign == 0)
    {
      /* Do the necessary preparations for a new frame */
      if (para_start_frame ())
      {
        /* Shellout gave SKIP_ONCE_RET -> search next block */
	para_assign_work ();
        return;
      };
    };
    slave_stat[task].frame_assigned = frame;
    para_send_work (frame_stat[frame].block_to_assign);
    return;
  };

  /* there is nothing left ro assign - stop slave */
  MPI_Send ((char *) NULL, 0, MPI_CHAR, task, PARA_KILL_SLAVE,
            MPI_COMM_WORLD);
  return;
}


/* para_display_plot
 *
 * displays a part from one line in the preview window
 */
static void para_display_plot (COLOUR* row_ptr, int row, int col,
                               int ncols)
{
  unsigned char Red, Green, Blue, Alpha;
  DBL grey;
  int i;

  /* draw Line */
  for (i = col; i < col + ncols; i++)
  {
    extract_colors (row_ptr[i], &Red, &Green, &Blue, &Alpha, &grey);
    POV_DISPLAY_PLOT (0, i, row, Red, Green, Blue, Alpha);
  };

  /* force draw of the whole line (with X11) */
  i = opts.Last_Column -1;
  extract_colors (row_ptr[i], &Red, &Green, &Blue, &Alpha, &grey);
  POV_DISPLAY_PLOT (0, i, row, Red, Green, Blue, Alpha);
}


/* para_receive_data
 *
 * one of the slaves sends a part of it's computed data
 */
static void para_receive_data ()
{
  int i, j;
  int size;
  int block_width, block_height;
  struct work_s work;
  COLOUR* colbuf;

  /* first receive block info */
  MPI_Recv (&work, 1, MPI_work_t, task, PARA_SEND_RESULTS,
            MPI_COMM_WORLD, &status);
#ifdef PARA_DEBUG
  Status_Info ("receiving block: %d of frame: "
               "%d, row %d-%d, col: %d-%d\n",
               work.block_num, work.frame, work.up, work.down,
               work.left, work.right);
#endif

  block_width = work.right - work.left;
  block_height = work.down - work.up;
  frame = work.frame;

  /* Make some checks to see if the data is not corrupted or invalid */
  if ((work.up    < opts.First_Line) ||
      (work.left  < opts.First_Column) ||
      (work.down  > opts.Last_Line)  ||
      (work.right > opts.Last_Column) ||
      (frame      < 0)               ||
      (frame      >= frames) ||
      (block_width  > para_chunk_width) ||
      (block_height > para_chunk_height))
  {
    Warning (0, "Bad block received frame from task: %d\n",
             task);
    Warning (0, "block: %d of frame: %d, row %d-%d, col: %d-%d\n",
               work.block_num, work.frame, work.up, work.down,
               work.left, work.right);
    return;
  };

  size = block_height * block_width * 5;
  colbuf = (COLOUR *) POV_MALLOC (size * sizeof (float), "COLOUR buffer");

  /* Now receive the pixels */
  MPI_Recv (colbuf, size, MPI_FLOAT, task,
            PARA_SEND_RESULTS, MPI_COMM_WORLD, &status);

  /* get one row after the other */
  for (i = work.up, j = 0; i < work.down; i++, j++)
  {
    /* Is incoming data current? */
    if (i >= frame_stat[frame].block_stat[work.block_num].wr_line)
    {
      if (frame_stat[frame].row_ptr[i] == NULL)
      {
        /* Starting a new row */
        frame_stat[frame].row_ptr[i] = (COLOUR *) POV_MALLOC
            (Frame.Screen_Width * sizeof (COLOUR), "row_ptr");
        while (frame_stat[frame].row_ptr[frame_stat[frame].
               row_to_malloc] != NULL &&
               frame_stat[frame].row_to_malloc < opts.Last_Line)
        {
          frame_stat[frame].row_to_malloc++;
        };
      };

      /* get color data */
      memcpy(frame_stat[frame].row_ptr[i] + work.left,
             colbuf + (j * block_width),
             (block_width) * sizeof (COLOUR));
      frame_stat[frame].block_stat[work.block_num].wr_line = i + 1;

      if (frame == frame_display)
      {
        para_display_plot (frame_stat[frame].row_ptr[i], i, work.left,
                           block_width);
      };

      /* Statistics - Attribute this grid section */
      grid_lines_done++;
      slave_stat[task].pixels_done += block_width;

      npixels_done += block_width;
    }
  };

  /* Has this entire block been rendered? */
  if (frame_stat[frame].block_stat[work.block_num].wr_line >=
      (work.block_num / grid_ppline + 1) * para_chunk_height +
      opts.First_Line - 1 ||
      frame_stat[frame].block_stat[work.block_num].wr_line >=
      opts.Last_Line)
  {
#ifdef PARA_DEBUG
    Status_Info ("Block: %d completed by task: %d\n",
                 work.block_num, task);
#endif
    frame_stat[frame].block_stat[work.block_num].status =
        PARA_GRID_POINT_DONE;
    slave_stat[task].blocks_done++;
  };

  if ((opts.Options & VERBOSE) &&
      (grid_lines_done % para_chunk_height) == 0)
  {
    para_dump_time (RENDER_STATS_FORMAT);
    Status_Info("%5.2f%% of blocks complete.", (float) 100 *
                grid_lines_done / para_chunk_height / grid_points /
                frames);
  };
}


/* para_write
 *
 * Writes out as many lines as possible in the output-file
 */
static void para_write ()
{
  int i;

  if (frame_stat[frame].block_to_write >= grid_points)
  {
    return;
  };

  do {
    /* check, if there is a complete line to write */
    for (i = frame_stat[frame].block_to_write;
         i < frame_stat[frame].block_to_write + grid_ppline; i++)
    {
      if (frame_stat[frame].block_stat[i].wr_line <=
          frame_stat[frame].row_to_write)
      {
#ifdef PARA_DEBUG
        Status_Info (" needing block %d in frame %d\n", i, frame);
#endif
        return;
      };
    };

    /* OK, we can write a whole line */
    if (opts.Options & DISKWRITE)
    {
#ifdef PARA_DEBUG
      Status_Info ("write: %d %d\n", frame,
                   frame_stat[frame].row_to_write);
      assert (frame_stat[frame].output_file != NULL);
#endif

      frame_stat[frame].output_file->Write_Line (
          frame_stat[frame].row_ptr[frame_stat[frame].row_to_write]);
    };

    /* Still lines to allocate so */
    if (frame_stat[frame].row_to_malloc < opts.Last_Line)
    {
      /* move this row buffer down, so we avoid one malloc/free */
      frame_stat[frame].row_ptr[frame_stat[frame].row_to_malloc] =
          frame_stat[frame].row_ptr[frame_stat[frame].row_to_write];
      frame_stat[frame].row_ptr[frame_stat[frame].row_to_write] = NULL;
      frame_stat[frame].row_to_malloc++;
    }
    /* otherwise free the row buffer */
    else
    {
      POV_FREE (frame_stat[frame].row_ptr
                [frame_stat[frame].row_to_write]);
      frame_stat[frame].row_ptr[frame_stat[frame].row_to_write] = NULL;
    };

    frame_stat[frame].row_to_write++;

#ifdef PARA_DEBUG
    if ((opts.Options & VERBOSE) && 
        (frame_stat[frame].row_to_write % 8) == 0)
    {
      para_dump_time (RENDER_STATS_FORMAT);
      Status_Info ("\r%5.2f%% of blocks complete.", (float) 100 *
                   grid_lines_done / para_chunk_height /grid_points /
                   frames);
      Status_Info (" %4d of %4d lines finished (in frame %d).",
                   frame_stat[frame].row_to_write - opts.First_Line,
                   opts.Last_Line - opts.First_Line, frame);

    };
#endif

    /* Finished a whole row of blocks */
    if ((frame_stat[frame].row_to_write - opts.First_Line) %
        para_chunk_height == 0)
    {
      for (i = frame_stat[frame].block_to_write;
           i < frame_stat[frame].block_to_write + grid_ppline; i++)
      {
        frame_stat[frame].block_stat[i].status = 
            PARA_GRID_POINT_WRITTEN;
      };
      frame_stat[frame].block_to_write += grid_ppline;
    };


  } while (frame_stat[frame].block_to_write < grid_points);
}


/* para_dump_time
 *
 * computes the elapsed time and prints it out.
 * the format of the output can be chosen by the flag parameter
 *
 * parameters:	flag	use this flag to chose the format of the output
 */
static void para_dump_time (int flag)
{
  DBL time_dif;
  unsigned long hrs, mins;
  DBL secs;

  STOP_TIME

  time_dif = TIME_ELAPSED

  if (time_dif > 0.0)
  {
    SPLIT_TIME (time_dif, &hrs, &mins, &secs);
    if (flag == TOTAL_STATS_FORMAT)
    {
      Statistics ("Total Time:   %3ld hours  %02ld minutes  "
                  "%02ld seconds", hrs, mins, (long)secs);
    }
    else if (flag == RENDER_STATS_FORMAT)
    {
      Status_Info ("\r%3ld:%02ld:%02ld ", hrs, mins, (long)secs);
    };
  }
  else
  {
    if (flag == TOTAL_STATS_FORMAT)
    {
      Statistics ("Total Time:   0 hours  0 minutes  0 seconds");
    }
    else if (flag == RENDER_STATS_FORMAT)
    {
      Status_Info("\r  -:--:-- ");
    };
  };
}


/* para_receive_stats
 *
 * receives a status msg and stores the data
 */
static void para_receive_stats ()
{
  struct stat_s stat;

  MPI_Recv (&stat, 1, MPI_stat_t, task,
            PARA_SEND_STATS, MPI_COMM_WORLD, &status);

  /* add COUNTER struct to total stats */
  sum_statistics (statistics, stat.stats);

  /* add render time to the current slave */
  slave_stat[task].trender += stat.trender;
}


/* para_print_stats
 *
 * prints out some info about the rendering,
 * the client tasks and the work they have done,
 * as well as the total time needet to execute parapov.
 */
static void para_print_stats ()
{
  int i;
  long n_pixels;

  n_pixels = (opts.Last_Line - opts.First_Line) *
         (opts.Last_Column - opts.First_Column);

  PRINT_STATS ("Standard statistics", statistics);
  Statistics ("ParaPov statistics:\n");
  Statistics ("--------------------------------------"
              "--------------------------------------\n");
  Statistics ("Number of Frames:\t\t%d\n", frames);
  if (frames > 1)
  {
    Statistics ("Pixels per Frame:\t\t%d\n", n_pixels);
  };
  Statistics ("Total Number of Pixels:\t\t%d\n", 
              (long) n_pixels * frames);
  Statistics ("--------------------------------------"
              "--------------------------------------\n");
  Statistics ("Block Height:\t\t\t%d\n", para_chunk_height);
  Statistics ("Block Width:\t\t\t%d\n", para_chunk_width);
  Statistics ("Blocks per Line:\t\t%d\n", grid_ppline);
  Statistics ("Blocks per Column:\t\t%d\n", grid_points / grid_ppline);
  if (frames > 1)
  {
    Statistics ("Blocks per Frame:\t\t%d\n", grid_points);
  };
  Statistics ("Total Number of Blocks:\t\t%d\n", grid_points * frames);
  Statistics ("--------------------------------------"
              "--------------------------------------\n");
  Statistics ("Number of Client Tasks:\t\t%d\n\n", ntasks - 1);
  for (int i = 1; i < ntasks; i++)
  {
    Statistics ("Task %d:  %s, %s, %s\n", i,
                slave_stat[i].info.nodename, slave_stat[i].info.machine,
                slave_stat[i].info.sysname);
    Statistics ("  Rendering Time:\t\t\t%6.3f secs\n",
                slave_stat[i].trender);
    Statistics ("  Time per Block:\t\t\t%6.3f secs\n",
                slave_stat[i].trender / slave_stat[i].blocks_done);
    Statistics ("  Pixels done:\t\t%5.2f%%\t\t%d\n",
                (DBL) slave_stat[i].pixels_done * 100 /
                npixels_done, slave_stat[i].pixels_done);
    Statistics ("  Blocks done:\t\t%5.2f%%\t\t%d\n",
                (DBL) slave_stat[i].blocks_done * 100 /
                grid_points / frames, slave_stat[i].blocks_done);
  };
  Statistics ("--------------------------------------"
              "--------------------------------------\n");

  para_dump_time (TOTAL_STATS_FORMAT);

  Statistics ("\n\n\n\n");
}


/* para_master_kill_slaves
 *
 * Kills all the slave tasks
 */
void para_master_kill_slaves ()
{
  int flag;
  MPI_Request dummy;

  for (int i = 1; i < ntasks; i++)
  {
#ifdef PARA_DEBUG
    stream_printf (DEBUG_STREAM, "Killing slave with task id = %d\n",i);
#endif
    
    /* use non blocking funktion. slave could already be down */
    MPI_Isend ((char *) NULL, 0, MPI_CHAR, i, PARA_KILL_SLAVE,
               MPI_COMM_WORLD, &dummy);
  };
}


/* para_master_init
 *
 * Initializes the server.
 * Sets all the block data, allocates memory
 * needed and inits the slave tasks.
 */
void para_master_init ()
{
  int i, j;
  struct info_s info;

  // sleep (1);
#ifdef PARA_DEBUG
  Status_Info ("initializing ParaPOV master...\n");
#endif
  
  /* set block data and check if user input is valid */
  para_chunk_height = opts.Para_Chunk_Height;
  para_chunk_width = opts.Para_Chunk_Width;
  
  if (para_chunk_height < 1 ||
      para_chunk_height > opts.Last_Line)
  {
    Status_Info ("\n\nWarning: "
                 "Invalid block height %d, using default %d\n",
                 para_chunk_height, DEFAULT_CHUNK_HEIGHT);
    para_chunk_height = DEFAULT_CHUNK_HEIGHT;
  };
  if (para_chunk_width < 1 ||
      para_chunk_width > opts.Last_Column)
  {
    Status_Info ("\n\nWarning: "
                 "Invalid block width %d, using default %d\n",
                 para_chunk_width, DEFAULT_CHUNK_WIDTH);
    para_chunk_width = DEFAULT_CHUNK_WIDTH;
  };

  grid_ppline = (opts.Last_Column - opts.First_Column +
                 para_chunk_width - 1) / para_chunk_width;
  grid_points =
      ((opts.Last_Line - opts.First_Line + para_chunk_height -1) /
        para_chunk_height) * grid_ppline;

#ifdef PARA_DEBUG
  Status_Info ("  mytid  = %d\n", mytid);
  Status_Info ("  ntasks = %d\n", ntasks);
  Status_Info ("  grid_ppline = %d\n", grid_ppline);
  Status_Info ("  grid_points = %d\n", grid_points);
#endif


  /* init frame_stat */
  frames = (opts.FrameSeq.FrameType == FT_MULTIPLE_FRAME) ?
      opts.FrameSeq.FinalFrame - opts.FrameSeq.InitialFrame + 1 : 1;
  frame_stat = (para_frame_stat *)
      POV_CALLOC (frames, sizeof (para_frame_stat), "frame_stat");
  frame_display = -1;

  for (i = 0; i < frames; i++)
  {
    frame_stat[i].status = PARA_FRAME_FREE;

    frame_stat[i].block_to_assign = 0;

    frame_stat[i].block_to_write = 0;

    frame_stat[i].row_to_write = opts.First_Line;
    frame_stat[i].row_to_malloc = opts.First_Line;

    frame_stat[i].block_stat = (para_block_stat *)
        POV_CALLOC (grid_points, sizeof (para_block_stat), "block_stat");

    for (j = 0; j < MaxStat; j++)
    {
      Init_Counter (frame_stat[i].stats[j]);
    };
  };


  /* init slave stat */
  slave_stat = (para_slave_stat *) 
      POV_CALLOC (ntasks,  sizeof (para_slave_stat), "slave_stat");
  for (i = 1; i < ntasks; i++)
  {
    MPI_Recv (&info, 1, MPI_info_t, i,
              PARA_SLAVE_CONNECTED, MPI_COMM_WORLD, &status);
    task = status.MPI_SOURCE;

    strcpy(slave_stat[task].info.sysname, info.sysname);
    strcpy(slave_stat[task].info.nodename, info.nodename);
    strcpy(slave_stat[task].info.machine, info.machine);
    strcpy(slave_stat[task].info.logname, info.logname);
#ifdef PARA_DEBUG
    Status_Info ("slave connected:\n");
    Status_Info ("\ttask id  = %d\n", task);
    Status_Info ("\tsysname  = %s\n", slave_stat[task].info.sysname);
    Status_Info ("\tnodename = %s\n", slave_stat[task].info.nodename);
    Status_Info ("\tmachine  = %s\n", slave_stat[task].info.machine);
    Status_Info ("\tlogname  = %s\n", slave_stat[task].info.logname);
#endif
    slave_stat[task].frame_assigned = PARA_SLAVE_FREE;
    slave_stat[task].pixels_done = 0;
    slave_stat[task].blocks_done = 0;
    slave_stat[task].trender = 0;
  };
  

  /* init statistics pointer and set to 0 */
  statistics = (COUNTER *)
      POV_CALLOC (MaxStat, sizeof (COUNTER), "statistics");
  init_statistics (statistics);
}


/* para_master_loop
 *
 * Handles slave requests, distributes and collects work.
 */
void para_master_loop ()
{
  int exit_val;
  struct error_s err;

  next_frame_finish = 0;

  while ((next_frame_finish < frames))
  {
    /* check for user abort */
    TEST_ABORT;
    if (Stop_Flag)
    {
      Render_Info ("\nAborting render...\n");
      Error ("User abort.\n");
      para_master_exit ();
      return;
    };

    /* check the message that will be send */
    MPI_Probe (MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);

    task = status.MPI_SOURCE;

#ifdef PARA_DEBUG
    Status_Info ("Master: slave %d sends msg %d, ", task,
                 status.MPI_TAG);
#endif

    switch (status.MPI_TAG)
    {
      case PARA_REQUEST_WORK:		/* a slave wants work		*/
#ifdef PARA_DEBUG
        Status_Info ("slave wants work.\n");
#endif
        MPI_Recv ((char *) NULL, 0, MPI_CHAR, task,
                  PARA_REQUEST_WORK, MPI_COMM_WORLD, &status);
        para_assign_work ();
        break;
      case PARA_SEND_RESULTS:		/* a slave sends data		*/
#ifdef PARA_DEBUG
        Status_Info ("slave sends results.\n");
#endif
        para_receive_data ();
        para_write ();
        para_finish_frame ();
        break;
      case PARA_SEND_STATS:		/* a slave sends status info	*/
#ifdef PARA_DEBUG
        Status_Info ("slave sends stats.\n");
#endif
        para_receive_stats ();
        break;
      case PARA_SLAVE_EXITED:
        MPI_Recv (&exit_val, 1, MPI_INT, task,
                  PARA_SLAVE_EXITED, MPI_COMM_WORLD, &status);
#ifdef PARA_DEBUG
        Status_Info ("Slave %d exited with exit value %d...\n",
                     task, exit_val);
#endif
        break;
      case PARA_ERROR:			/* a slave sends a error msg	*/
#ifdef PARA_DEBUG
        Status_Info ("slave sends error.\n");
#endif
        MPI_Recv (&err, 1, MPI_error_t, task,
                  PARA_ERROR, MPI_COMM_WORLD, &status);
        Status_Info ("\n\nERROR received from task %d:\n", task);
        Status_Info ("\tmpi_error = %d\n", err.mpi_error);
        Status_Info ("\tmsg       = %s\n\n", err.msg);
        break;
      default:
        Error ("unexpected message received: %d\n", status.MPI_TAG);
    };
  };
  
  para_master_exit ();
}
