/* Slave interface implementation
 *
 * File: parapov_slv.cpp	Author: Matthias Fueller
 * Date: 13.11.2002 		Version: 0.1
 *
 * Change log:
 */
#include <stdlib.h>
#include <stdio.h>
#include <sys/utsname.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include "frame.h"
#include "pov_mem.h"
#include "userio.h"
#include "povray.h"
#include "povproto.h"
#include "parapov.h"
#include "parapov_slv.h"
#include "colutils.h"


/* definition of global variables */
int para_next_frame;
DBL para_render_time;
BufferImage *para_buffer_singleton;

/* static variables */
static int frame;       		/* current frame 		*/
static int block;                      	/* current block		*/
static char logname [INFO_NAME_LEN];	/* name of log file		*/
static FILE *fileout;			/* log file			*/
static unsigned char first_frame;	/* flag for first frame 	*/

/* static functions */
static void para_file_console (char *s);


/* para_redirect_slave_output
 *
 * All slave output will be sent to a file called n{mytid}.log,
 * this filename is unique for each task.
 */
void para_redirect_slave_output ()
{
  int a;

  sprintf (logname, "parapov_n%d.log", mytid);
  fileout = fopen (logname, "w");
  if (fileout)
  {
    for (a = 0; a < MAX_STREAMS; a++)
    {
      if (Stream_Info [a].console)
      {
        Stream_Info [a].console = para_file_console;
      };
    };
    Status_Info ("This is the logfile of task #%d\n\n", mytid);
  };
}


/* para_client_init
 *
 * Initializes the client
 */
void para_slave_init ()
{
  struct utsname uname_info;
  struct info_s  info;
  int	 niceInc, err;

  first_frame = 1;

  /* renice this process - dont consume too much cpu time */
  niceInc = opts.Para_Slave_Niceness;
  if (opts.Para_Slave_Niceness != 0)
  {
    err = nice (niceInc);
    if (err)
    {
      Warning (0, "Could not set priority: %s\n", strerror (errno));
    };
  };

  if (opts.Para_Slave_Logging)
  {
    para_redirect_slave_output ();
  };

  /* get information about the system */
  uname (&uname_info);
  strcpy (info.sysname, uname_info.sysname);
  strcpy (info.nodename, uname_info.nodename);
  strcpy (info.machine, uname_info.machine);

  strcpy (info.logname, logname);

  /* Inform master that we are connected and send some info */
  MPI_Send (&info, 1, MPI_info_t, 0, PARA_SLAVE_CONNECTED,
  	    MPI_COMM_WORLD);

#ifdef PARA_DEBUG
  Debug_Info ("sent info to master\n");
#endif

  /* Slave must be able to write to disk - we will redirect it and
   * send it to the master
   */
  opts.Options |= DISKWRITE;

  /* turn off display */
  opts.Options &= ~DISPLAY;

  frame = -1;
  para_next_frame = -1;

  /* construct the only image buffer object we have */
  para_buffer_singleton = POV_NEW (BufferImage);
}


/* para_slave_get_work
 *
 * looks if there are work messages
 *
 * The end coordinates (work.down and work.right) are not rendered,
 * the last line rendered is work.down - 1, the last column is
 * work.right - 1
 */
slave_status para_slave_get_work ()
{
  MPI_Status	stat;
  struct work_s work;
  int		flag;

  /* check if there is a PARA_KILL_SLAVE message waiting */
  MPI_Iprobe(0, PARA_KILL_SLAVE, MPI_COMM_WORLD, &flag, &stat);

  /* Request work unless there is a PARA_KILL_SLAVE msg */
  if(!flag)
  {
    MPI_Send ((char *) NULL, 0, MPI_CHAR, 0, PARA_REQUEST_WORK,
  	      MPI_COMM_WORLD);

#ifdef PARA_DEBUG
    Debug_Info ("Task %d: work requested...\n", mytid);
#endif
  };

  work.frame = -1;
  MPI_Recv (&work, 1, MPI_work_t, 0, MPI_ANY_TAG, MPI_COMM_WORLD,
  	    &stat);

  /* analyze the message */
  switch (stat.MPI_TAG)
  {
    case PARA_KILL_SLAVE:		/* slave has to exit */
#ifdef PARA_DEBUG
      Debug_Info ("Slave %d: KILL SLAVE RECEIVED, exiting!!\n", mytid);
#endif
      
      para_next_frame = opts.FrameSeq.FinalFrame + 1;
      return ST_EXIT;

    case PARA_SEND_WORK:		/* new block received */
#ifdef PARA_DEBUG
      Debug_Info ("Work received: frame %d, block %d\n"
      		  "left %d, up %d, right %d, down %d\n", work.frame,
                  work.block_num, work.left, work.up, work.right,
                  work.down);
#endif
      block = work.block_num;

      /* Tell the renderer what to trace */
      opts.First_Line = work.up;
      opts.Last_Line = work.down;
      opts.First_Column = work.left;
      opts.Last_Column = work.right;

      /* configure the buffer for use with the current block */
      para_buffer_singleton->init ();

      if (frame != work.frame)		/* new frame started 		*/
      {
#ifdef PARA_DEBUG
	Debug_Info ("New Frame started\n");
#endif

        if (!first_frame)
	{
	  para_slave_send_stats ();
	};
	first_frame = 0;

        frame = work.frame;
	para_next_frame = opts.FrameSeq.InitialFrame + frame;
	return ST_NEW_FRAME;
      };

      /* initialize povray for rendering */
      Initialize_Renderer ();

      return ST_NEW_BLOCK;

    default:
      Error ("Unknown MPI message &d\n", stat.MPI_TAG);
  };

  /* there occured an error if we reach here */
  para_next_frame = opts.FrameSeq.FinalFrame + 1;
  return ST_EXIT;
}



/* para_slave_exit
 *
 * Exits with exit code err after sending a PARA_SLAVE_EXITED
 * msg to the master.
 */
void para_slave_exit (int err)
{
  MPI_Request dummy;

  if (frame != -1)
  {
    para_slave_send_stats ();
  };

  /* Synchronize with master. The master must wait until
   * the slave has send the statistics. The corresponding
   * MPI_Barrier in the master code is in para_master_exi
   */
  MPI_Barrier (MPI_COMM_WORLD);

  MPI_Isend (&err, 1, MPI_INT, 0, PARA_SLAVE_EXITED,
	     MPI_COMM_WORLD, &dummy);

  if (para_buffer_singleton)
  {
    POV_DELETE (para_buffer_singleton, BufferImage);
    para_buffer_singleton = NULL;
    Output_File = NULL;
  };

#ifdef PARA_DEBUG
  Debug_Info ("Task %d says good bye\n", mytid);
#endif

  /* Synchronize with master. We must wait until the master
   * received our statistics. Corresponding MPI_Barrier in the
   * Master code is in para_master_exit.
   */
  MPI_Barrier (MPI_COMM_WORLD);

  /* shut down MPI */
  para_close ();

  /* flush and close the log file */
  fflush (fileout);
  fclose (fileout);


  EXIT_POVRAY (err);
}


/* para_slave_send_error
 *
 * Send an error msg to the server
 */
void para_slave_send_error (const char *msg, int mpi_error)
{
  struct error_s err;

#ifdef PARA_DEBUG
  Debug_Info ("Send error to master: %s\n", msg);
#endif

  err.mpi_error = mpi_error;
  if (strlen (msg) < ERR_BUF_SIZE - 1)
  {
    strcpy (err.msg, msg);
  }
  else					/* cut the message 		*/
  {
    strncpy (err.msg, msg, ERR_BUF_SIZE - 1);
    err.msg [ERR_BUF_SIZE - 1] = 0;		
  };

  MPI_Send (&err, 1, MPI_error_t, 0, PARA_ERROR, MPI_COMM_WORLD);
}


/* para_file_console
 *
 * Writes text to a file, used for output redirection
 */
static void para_file_console (char *s)
{
  if (fileout)
  {
    fputs (s, fileout);
  }
  else
  {
    fputs (s, stdout);
  };
}


/* para_slave_send_stats
 *
 * Send the statistics of this frame to the master
 */
void para_slave_send_stats ()
{
  struct stat_s mystats;

  mystats.frame = frame;
  mystats.trender = para_render_time;
  memcpy (&mystats.stats, &stats, sizeof (COUNTER) * MaxStat);

#ifdef PARA_DEBUG
  Debug_Info ("Task %d: sending stats for frame %d ...\n", mytid,
	      frame);
#endif

  MPI_Send (&mystats, 1, MPI_stat_t, 0, PARA_SEND_STATS,
  	    MPI_COMM_WORLD);
}



/*
 * BufferImage implementation
 */


/* BufferImage constructor
 *
 * Initializes the variables
 */
BufferImage::BufferImage ()
{
  buf = NULL;
  valid = 0;
}


/* BufferImage destructor
 *
 * Releases the buffer
 */
BufferImage::~BufferImage ()
{
#ifdef PARA_DEBUG
  if ((valid) && (line < height))
  {
    Debug_Info ("Closing buffer with %d lines not sent\n",
    		height - line);
  };
#endif

  if (buf)
  {
    POV_FREE (buf);
  };
}


/* BufferImage::init
 *
 * Initializes the object for use with the current block,
 * must be called if the block changes
 */
void BufferImage::init ()
{
  if (buf)
  {
#ifdef PARA_DEBUG
    Debug_Info("deleting old buffer in BufferImage::init\n");
#endif
    POV_FREE (buf);
  };

  this->width = opts.Last_Column - opts.First_Column;
  this->height = opts.Last_Line - opts.First_Line;
  line = 0;
  last_line = 0;

  buf = (COLOUR *) POV_CALLOC (sizeof (COLOUR), width * height,
  			       "colbuf");
  lines_to_send = (opts.Last_Line - opts.First_Line + 3) / 4;

#ifdef PARA_DEBUG
  Debug_Info ("created buffer with width: %d height: %d "
  	      "lines to send: %d\n", width, height, lines_to_send);
#endif

  valid = 1;
}


/* BufferImage::Write_Line
 *
 * Write one line to the buffer.
 * If enough lines are complete or if the end of the block is reached,
 * then lines are sent to the server.
 */
void BufferImage::Write_Line (COLOUR *line_data)
{
  struct work_s work;
  int 		offset;
  COLOUR	*block_begin;

  if (!valid)
  {
    Error ("Attempt to write to invalid block, current block is %d\n",
           block);
  };

#ifdef PARA_DEBUG
  Debug_Info ("in Write_Line: lines %d , last_line %d\n",
  	      line, last_line);
#endif
#ifdef PARA_DEBUG_PIXEL
  int a;
  unsigned char red, green, blue, alpha;
  double gray;

  for(a = 0; a < width; a++)
  {
  	extract_colors(line_data[a], &red, &green, &blue, &alpha, &gray);
  	Debug_Info ("%d, %d, %d\t", red, green, blue);
  };
#endif


  offset = line * width;
  memcpy (buf + offset, line_data + opts.First_Column,
  	  width * sizeof (COLOUR));

  line++;

  /* Send lines if we have lines_to_send lines complete
   * or if we reached the end of the block
   */
  assert (lines_to_send != 0);
  if (((line % lines_to_send) == 0) ||
      (line >= height))
  {
    work.frame = frame;
    work.block_num = block;

    work.left = opts.First_Column;
    work.right = opts.Last_Column;
    work.up = opts.First_Line + last_line;
    work.down = opts.First_Line + line;
#ifdef PARA_DEBUG
    Debug_Info ("sending results to master: frame %d, block %d, "
    		"left %d, up %d, right %d, down %d\n",
    		work.frame, work.block_num,
		work.left, work.up, work.right, work.down);
#endif
    /* first send block info */
    MPI_Send (&work, 1, MPI_work_t, 0, PARA_SEND_RESULTS,
  	      MPI_COMM_WORLD);

    block_begin = buf + (last_line * width);

    /* now send the pixels */
    MPI_Send (block_begin,
  	      (line - last_line) * width * 5,
	      MPI_FLOAT, 0, PARA_SEND_RESULTS, MPI_COMM_WORLD);

    last_line = line;

    /* block done, reset */
    if (line >= height)
    {
      POV_FREE (buf);
      buf = NULL;
      valid = 0;
    };
  };
}


/* BufferImage::Read_Line
 *
 * Read one line from the image.
 * NOT IMPLEMENTED yet, this is used for render continuation, which
 * is not supported by parapov.
 */
int BufferImage::Read_Line (COLOUR *line_data)
{
  Error ("Read_Line in BufferImage called, current block is %d\n",
         block);

  return 0;
}



